@@ -1,13 +1,13 @@
 import logging
 
+from abnosql.kms import kms
 from abnosql.table import table
 
-__version__ = '0.0.1'
-
 
 logger = logging.getLogger(__name__)
 logger.addHandler(logging.NullHandler())
 
 __all__ = [  # type: ignore
+    kms,
     table
 ]
