@@ -6,7 +6,12 @@ import pluggy  # type: ignore
 
 import abnosql.exceptions as ex
 from abnosql.plugin import PM
+from abnosql.table import add_audit
 from abnosql.table import get_sql_params
+from abnosql.table import kms_decrypt_item
+from abnosql.table import kms_encrypt_item
+from abnosql.table import kms_process_query_items
+from abnosql.table import parse_connstr
 from abnosql.table import TableBase
 from abnosql.table import validate_query_attrs
 
@@ -86,13 +91,23 @@ class Table(TableBase):
         _client = self.config.get('database_client', self.database_client)
         if _client is not None:
             return _client
+        pc = parse_connstr()
         cf = {}
+        if pc is not None:
+            cf.update({
+                'account': pc.username,
+                'database': pc.hostname,
+                'credential': pc.password
+            })
         required = ['endpoint', 'credential', 'database']
         for attr in ['account'] + required:
-            cf[attr] = self.config.get(
+            val = self.config.get(
                 attr, os.environ.get('ABNOSQL_COSMOS_' + attr.upper())
             )
-        if cf['endpoint'] is None and cf['account'] is not None:
+            # override
+            if val is not None:
+                cf[attr] = val
+        if cf.get('endpoint') is None and cf['account'] is not None:
             cf['endpoint'] = 'https://%s.documents.azure.com' % cf['account']
         missing = [_ for _ in required if cf[_] is None]
         if len(missing):
@@ -115,18 +130,33 @@ class Table(TableBase):
         _item = self.pm.hook.get_item_post(table=self.name, item=item)
         if _item:
             item = _item
+        item = kms_decrypt_item(self.config, item)
         return item
 
     @cosmos_ex_handler()
-    def put_item(self, item: t.Dict):
+    def put_item(
+        self,
+        item: t.Dict,
+        user: t.Optional[str] = None
+    ):
+        if user:
+            item = add_audit(item, user)
+        _item = self.pm.hook.put_item_pre(table=self.name, item=item)
+        if _item:
+            item = _item[0]
+        item = kms_encrypt_item(self.config, item)
         self._container(self.name).upsert_item(item)
         self.pm.hook.put_item_post(table=self.name, item=item)
 
     @cosmos_ex_handler()
-    def put_items(self, items: t.Iterable[t.Dict]):
+    def put_items(
+        self,
+        items: t.Iterable[t.Dict],
+        user: t.Optional[str] = None
+    ):
         # TODO(batch)
         for item in items:
-            self.put_item(item)
+            self.put_item(item, user)
         self.pm.hook.put_items_post(table=self.name, items=items)
 
     @cosmos_ex_handler()
@@ -139,12 +169,13 @@ class Table(TableBase):
     @cosmos_ex_handler()
     def query(
         self,
-        key: t.Dict[str, t.Any],
+        key: t.Optional[t.Dict[str, t.Any]] = None,
         filters: t.Optional[t.Dict[str, t.Any]] = None,
         limit: t.Optional[int] = None,
         next: t.Optional[str] = None
     ) -> t.Dict[str, t.Any]:
         filters = filters or {}
+        key = key or {}
         validate_query_attrs(key, filters)
         parameters = {
             f'@{k}': v
@@ -160,12 +191,14 @@ class Table(TableBase):
             statement += f' {op} {self.name}.{param[1:]} = {param}'
             op = 'AND'
 
-        return self.query_sql(
+        resp = self.query_sql(
             statement,
             parameters,
             limit=limit,
             next=next
         )
+        resp['items'] = kms_process_query_items(self.config, resp['items'])
+        return resp
 
     @cosmos_ex_handler()
     def query_sql(
@@ -190,23 +223,47 @@ class Table(TableBase):
         }
         if len(params):
             kwargs['parameters'] = params
-        if limit:
-            kwargs['max_item_count'] = limit
-        # TODO(x-ms-continuation)
-        # from microsoft / planetary-computer-tasks on github
-        # The Python SDK does not support continuation tokens
-        # for cross-partition queries.
-        #
-        # response_hook callable doesnt show x-ms-continuation
-        # header with or without enable_cross_partition_query
-        # even when max_item_count = 1
-        # print(f'KWARGS: {kwargs}')
+        # python cosmos SDK doesnt support
+        # kwargs['max_item_count'] = limit
+        limit = limit or 100
+        next = next or '0'
+
+        # Python SDK does not support continuation
+        # for cross-partition queries - see limitations https://github.com/Azure/azure-sdk-for-python/tree/main/sdk/cosmos/azure-cosmos  # noqa
+        # so add OFFSET and LIMIT if not already present
+        if (
+            ' OFFSET ' not in statement.upper()
+            and ' LIMIT ' not in statement.upper()
+        ):
+            kwargs['query'] += f' OFFSET {next} LIMIT {limit}'
         container = self._container(self.name)
         items = list(container.query_items(**kwargs))
+        headers = container.client_connection.last_response_headers
+        # continuation = headers.get('x-ms-continuation')
+        # total size in 'x-ms-resource-usage' eg ;documentsCount=3
+        resource_usage = {
+            _.split('=', 1)[0]: _.split('=', 1)[1]
+            for _ in headers.get('x-ms-resource-usage', '').split(';')
+            if '=' in _
+        }
+        doc_count = None
+        try:
+            doc_count = int(resource_usage['documentsCount'])
+        except Exception:
+            doc_count = None
+
         for i in range(len(items)):
             items[i] = strip_cosmos_attrs(items[i])
+        items = kms_process_query_items(self.config, items)
+        _next = None
+        try:
+            _next = limit + int(next)
+        except Exception:
+            _next = None
+        if doc_count is not None and _next is not None and _next >= doc_count:
+            _next = None
 
         return {
             'items': items,
-            'next': None
+            'next': str(_next) if _next and len(items) else None
         }
