@@ -5,10 +5,16 @@ import typing as t
 import pluggy  # type: ignore
 import sqlglot  # type: ignore
 from sqlglot.executor import execute  # type: ignore
+from sqlglot import exp  # type: ignore
+from sqlglot import parse_one  # type: ignore
 
 import abnosql.exceptions as ex
 from abnosql.plugin import PM
+from abnosql.table import add_audit
 from abnosql.table import get_sql_params
+from abnosql.table import kms_decrypt_item
+from abnosql.table import kms_encrypt_item
+from abnosql.table import kms_process_query_items
 from abnosql.table import quote_str
 from abnosql.table import TableBase
 from abnosql.table import validate_query_attrs
@@ -39,6 +45,11 @@ def get_table_name(statement: str):
     ))
 
 
+def get_table_count(name: str) -> int:
+    global TABLES
+    return len(list(TABLES.get(name, {}).values()))
+
+
 def query_items(
     statement: str,
     items: t.List[t.Dict[str, t.Any]],
@@ -69,15 +80,34 @@ def query_items(
         table_name = get_table_name(statement)
 
     # sqlglot execute can't handle dict or list keys...
+    _items = []
     _unpack = {}
-    for i in range(len(items)):
-        for k, v in items[i].items():
+    for item in items:
+        new_item = item.copy()
+        for k in item.keys():
+            v = item[k]
             if type(v) in [dict, list]:
                 _unpack[k] = True
-                items[i][k] = json.dumps(v)
+                new_item[k] = json.dumps(v)
+        _items.append(new_item)
+
+    # get any offset supplied
+    offset = None
+    for _offset in parse_one(statement).find_all(exp.Offset):  # type: ignore
+        offset = _offset
+        break
+    if offset:
+        try:
+            offset = int(str(offset).split(' ')[-1])  # type: ignore
+        except Exception:
+            offset = None
+
+    # slice the data via offset
+    if isinstance(offset, int) and offset > 0 and offset < len(_items):
+        _items = _items[offset:]
 
     # query the data
-    resp = execute(statement, tables={table_name: items})
+    resp = execute(statement, tables={table_name: _items})
     rows = [
         dict(zip(resp.columns, row))
         for row in resp.rows
@@ -142,11 +172,19 @@ class Table(TableBase):
         _item = self.pm.hook.get_item_post(table=self.name, item=item)
         if _item:
             item = _item
+        item = kms_decrypt_item(self.config, item)
         return item
 
     @memory_ex_handler()
-    def put_item(self, item: t.Dict):
+    def put_item(
+        self,
+        item: t.Dict,
+        user: t.Optional[str] = None
+    ):
+        if user:
+            item = add_audit(item, user)
         key = ':'.join([item[_] for _ in self.key_attrs])
+        item = kms_encrypt_item(self.config, item)
         if self.items:
             self.items[key] = item
         else:
@@ -157,9 +195,13 @@ class Table(TableBase):
         self.pm.hook.put_item_post(table=self.name, item=item)
 
     @memory_ex_handler()
-    def put_items(self, items: t.Iterable[t.Dict]):
+    def put_items(
+        self,
+        items: t.Iterable[t.Dict],
+        user: t.Optional[str] = None
+    ):
         for item in items:
-            self.put_item(item)
+            self.put_item(item, user)
         self.pm.hook.put_items_post(table=self.name, items=items)
 
     @memory_ex_handler()
@@ -196,6 +238,7 @@ class Table(TableBase):
             statement += f' {op} {self.name}.{param[1:]} = {param}'
             op = 'AND'
         items = self.query_sql(statement)
+        items = kms_process_query_items(self.config, items)
         return {
             'items': items,
             'next': None
@@ -223,4 +266,6 @@ class Table(TableBase):
         else:
             global TABLES
             items = list(TABLES.get(self.name, {}).values())
-        return query_items(statement, items, params, self.name)
+        items = query_items(statement, items, params, self.name)
+        items = kms_process_query_items(self.config, items)
+        return items
